# -*- coding: utf-8 -*-


from rqlalchemy.query import RQLQueryMixIn, RQLQueryError


__title__ = "rqlalchemy"
__version__ = "0.4.1"
__author__ = "Pedro Werneck"
__license__ = "MIT"


__all__ = ["RQLQueryMixIn", "RQLQueryError"]
